
#include <PalmOS.h>
#include <BtLib.h>
#include <BtLibTypes.h>
#include <unix_stdarg.h>
#include <StringMgr.h>
#include <Event.h>
#include <SystemMgr.h>

#include "BtLibTester_Rsc.h"
//----------------------------------------------------------------

// {99ABEE90-122D-4242-9E8E-F47F7C5365E3}
static const BtLibSdpUuidType kL2capUUID = {btLibUuidSize128, 
{ 0x99, 0xab, 0xee, 0x90, 0x12, 0x2d, 0x42, 0x42, 0x9e, 0x8e, 
0xf4, 0x7f, 0x7c, 0x53, 0x65, 0xe3 }};

// {7D97B361-6221-4459-9241-227576CC6AB8}
static const BtLibSdpUuidType kRfCommUUID = {btLibUuidSize128,
{ 0x7d, 0x97, 0xb3, 0x61, 0x62, 0x21, 0x44, 0x59, 0x92, 0x41, 
0x22, 0x75, 0x76, 0xcc, 0x6a, 0xb8 }};

#define kAppFileCreator   'btlt'
#define kAppBtName  "BtLibTest"

#define kMaxSizeLog 4096
#define kBdAddrStringSize  18
#define kL2capSocketMtu   100

#define INVALID_SOCKET   0xFFFF
#define INVALID_SDP_RECORD 0

#define UNUSED(x) (x=x) // To avoid compiler warning about unused arguments
//----------------------------------------------------------------

UInt16 gBtLibRefNum = 0;

BtLibDeviceAddressType gBdAddr;
Char  gBdAddrStr[kBdAddrStringSize];
Char gDataStr[kL2capSocketMtu];

BtLibSocketRef gListenerSocket = INVALID_SOCKET;
BtLibSocketRef gConnectionSocket = INVALID_SOCKET;

BtLibSocketRef _sdpSocket = INVALID_SOCKET;
BtLibSocketRef _controlSocket = INVALID_SOCKET;
BtLibSocketRef _interruptSocket = INVALID_SOCKET;

BtLibSocketRef gL2CapSocket = INVALID_SOCKET;

const UInt32 gBufferLength = 256;
UInt8 gBuffer[gBufferLength];

BtLibSdpRecordHandle gSdpRecordH = INVALID_SDP_RECORD;
BtLibSdpAttributeDataType attributeValues;

Boolean  gLibOpened = false;
Boolean  gCanDraw = true;
//----------------------------------------------------------------

static void hidControlCallback(BtLibSocketEventType *sEventP, UInt32 refCon);
static void hidInterruptCallback(BtLibSocketEventType *sEventP, UInt32 refCon);
static void btManagementCallback(BtLibManagementEventType *mEventP, UInt32 refCon);

static void ChangeLabel(UInt16 labelID, const Char *newLabel);
static void UpdateScrollbar();
static void ScrollLines(UInt16 numLinesToScroll, WinDirectionType direction, Boolean update);
static void logPrintf(const Char *format, ...);
static void logPrintErr(Err e);
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: MainFormDoCommand
//
// DESCRIPTION: This routine performs the menu command specified.
//
// PARAMETERS:  command  - menu item id
//
// RETURNED: nothing
//
//----------------------------------------------------------------
static Boolean MainFormDoCommand(UInt16 command)
{
	Boolean handled = false;
	Err error = 0;

	return (handled);
}

//----------------------------------------------------------------
//
// FUNCTION: MainFormHandleEvent
//
// DESCRIPTION: This routine is the event handler for the 
//  "MainForm" of this application.
//
// PARAMETERS:  eventP  - a pointer to an EventType structure
//
// RETURNED: true if the event has handle and should not be passed
//  to a higher level handler.
//
//----------------------------------------------------------------
static Boolean MainFormHandleEvent(EventPtr eventP)
{
	switch (eventP->eType) 
    {
	    case menuEvent:
	        return MainFormDoCommand(eventP->data.menu.itemID);

	    case frmOpenEvent: 
        {
	        FormPtr frmP = FrmGetActiveForm();
	        Err error;

	        // Draw the form
	        FrmDrawForm ( frmP);
	        ChangeLabel(MainAddrLabel, "No Connection");
	        MemSet(&gBdAddr, sizeof(gBdAddr), 0);
	        UpdateScrollbar();

	        // Find the Library
	        if (SysLibFind(btLibName, &gBtLibRefNum) == 0)
            {
            	logPrintf("BtLib Found");
            }
	        else
            {
	            // load it if it can't be found 
	            if(SysLibLoad(sysFileTLibrary, sysFileCBtLib, &gBtLibRefNum) == 0)
                {
                	logPrintf("BtLib Loaded");
                }
	            else
                {
	                logPrintf("Could not find BtLib");
	                return true;
                }
            }
            
	        // Open the library
	        error = BtLibOpen(gBtLibRefNum, false);
	        if ((error != btLibErrNoError) && (error != btLibErrAlreadyOpen))
            {
	            logPrintf("BtLib Opened Failed: 0x%hx", error);
	            return true;
            }

	        gLibOpened = true;
			logPrintf("BtLib Opened");

	        // Register to receive bt management events.
	        BtLibRegisterManagementNotification(gBtLibRefNum, btManagementCallback, 0);
        
	        // Prevent drawing while the dialog is up.
	        gCanDraw = false;
        
	        // Discover a device
	        error = BtLibDiscoverSingleDevice(gBtLibRefNum, NULL, NULL, 0, &gBdAddr, false, false);
	        if (error == btLibErrCanceled)
            {
    	        logPrintf("Discovery Canceled");
	            return true;
            }
        
	        // Enable drawing again.
	        gCanDraw = true;
        
	        // Create ACL link.
	        BtLibAddrBtdToA(gBtLibRefNum, &gBdAddr, gBdAddrStr, kBdAddrStringSize);
	        logPrintf("Creating Link to %s...", gBdAddrStr);
	        error = BtLibLinkConnect(gBtLibRefNum, &gBdAddr);
	        if(error == btLibErrPending)
            {
	            logPrintf("Link Create Error: 0x%hx", error);
	            logPrintErr(error);
	            return true;
            }

	        // Only need this if we're using SDP protocol.
	        // Allocate a service record.
	        //Err error = BtLibSdpServiceRecordCreate(gBtLibRefNum, &gSdpRecordH);
	        //if (error != btLibErrNoError)
	        //    {
	        //    logPrintf("BtLibSdpServiceRecordCreate() failed (%i).", error);
	        //    logPrintErr(error);
	        //    }
	        return true;
        }

    case frmCloseEvent:
        {
	        Err err;

	        // A host or device shall always complete the disconnection of the interrupt channel before disconnecting the control channel.
	        if (_controlSocket != INVALID_SOCKET)
            {
	            logPrintf("Closing control socket.");
	            BtLibSocketClose(gBtLibRefNum, _controlSocket);
	            _controlSocket = INVALID_SOCKET;
            }
	        if (_interruptSocket != INVALID_SOCKET)
            {
	            logPrintf("Closing interrupt socket.");
	            BtLibSocketClose(gBtLibRefNum, _interruptSocket);
    	        _interruptSocket = INVALID_SOCKET;
            }
        
	        // Only need this if we're using SDP protocol.
	        //BtLibSdpServiceRecordDestroy(gBtLibRefNum, BtLibSdpRecordHandle recordH);
        
	        logPrintf("Disconnecting ACL link.");
	        BtLibLinkDisconnect(gBtLibRefNum, &gBdAddr);
        
	        // Close the library
	        logPrintf("Closing BtLib.");
	        if(gLibOpened)
            {
	            BtLibUnregisterManagementNotification(gBtLibRefNum, btManagementCallback);
	            err = BtLibClose(gBtLibRefNum);
            }
        
			FormPtr frmP = FrmGetFormPtr(MainForm);
			if (frmP)
			{
				Char *txtP = FldGetTextPtr((const FieldType*)FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainLogField)));
				if (txtP)
				{
					MemPtrFree(txtP);
				}
			}

			SysTaskDelay(SysTicksPerSecond()*2);

			return (true);
        }

    case sclRepeatEvent:
        {
	        Int16 val = eventP->data.sclRepeat.newValue - eventP->data.sclRepeat.value;
	        if (val > 0)
	            ScrollLines((UInt16)val,  winDown, false);
	        else
	            ScrollLines((UInt16)-val, winUp,   false);
        
	        return true;
        }

    case keyDownEvent:
        {
	        switch (eventP->data.keyDown.chr)
            {
            case pageUpChr:
                {
                ScrollLines(0, winUp, true);
                return true;
                }
            case pageDownChr:
                {
                ScrollLines(0, winDown, true);
                return true;
                }
            default:
                {
                logPrintf("KeyDown: %C", eventP->data.keyDown.chr);
                return (true);
                }
            }
        }

    case keyUpEvent:
        {
        logPrintf("KeyUp: %C", eventP->data.keyUp.chr);
        return (true);
        }
    }

return (false);
}

//----------------------------------------------------------------
//
// FUNCTION: AppHandleEvent
//
// DESCRIPTION: This routine loads form resources and set the event
//  handler for the form loaded.
//
// PARAMETERS:  event  - a pointer to an EventType structure
//
// RETURNED: true if the event was handled and should not be passed
//  to a higher level handler.
//
//----------------------------------------------------------------
static Boolean AppHandleEvent(EventPtr eventP)
{
	if (eventP->eType == frmLoadEvent)
	{
		UInt16 formId;
		FormPtr formP;

		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		formP = FrmInitForm(formId);
		FrmSetActiveForm(formP);

		// And set its event handler
		if (formId == MainForm)
			FrmSetEventHandler(formP, MainFormHandleEvent);
		return true;
	}
	else
	{
		return false;
	}
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION:  AppStart
//
// DESCRIPTION:  Initialize the application. Goto the main form.
//
// PARAMETERS:   Nothing
//
// RETURNED:  Nothing
//
//----------------------------------------------------------------
static void AppStart()
{
	FrmGotoForm(MainForm);
}

//----------------------------------------------------------------
//
// FUNCTION: AppStop
//
// DESCRIPTION: Terminate the application. Close the library.
//
// PARAMETERS:  Nothing
//
// RETURNED: Nothing
//
//----------------------------------------------------------------
static void AppStop()
{
	Err err = 0;

	// Close all the opened forms.
	FrmCloseAllForms ();
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: AppEventLoop
//
// DESCRIPTION: This routine is the event loop for the application.  
//
// PARAMETERS:  nothing
//
// RETURNED: nothing
//
//----------------------------------------------------------------
static void AppEventLoop()
{
	UInt16 error;
	EventType event;

	EvtGetEvent(&event, evtWaitForever);
	while (event.eType != appStopEvent)
	{
		// Enable drawing if the main form is active, disable if another
		// form or menu is opened on top
		if ((event.eType == winExitEvent) &&
			(event.data.winExit.exitWindow == (WinHandle)FrmGetFormPtr(MainForm)))
		{
			gCanDraw = false;
		}
		else if ((event.eType == winEnterEvent) &&
			(event.data.winEnter.enterWindow == (WinHandle)FrmGetFormPtr(MainForm)))
		{
			gCanDraw = true;
			logPrintf(NULL);
		}

		if ((!SysHandleEvent(&event)) &&
			(!MenuHandleEvent(0, &event, &error)) &&
			(!AppHandleEvent(&event)))
		{
			FrmDispatchEvent(&event);
		}

		EvtGetEvent(&event, evtWaitForever);
	}
}
//----------------------------------------------------------------

void hidControlCallback(BtLibSocketEventType *sEvent, UInt32 refCon)
{
	static Boolean bModeSet = false;
	if (sEvent->event == btLibSocketEventConnectedOutbound)
    {
		// Set LED pattern.
		UInt8 ledBuffer[] = {0x52, 0x11, 0xa0};
		BtLibSocketSend(gBtLibRefNum, _controlSocket, (UInt8*)&ledBuffer, sizeof(ledBuffer));

		Err error;
		BtLibSocketCreate(gBtLibRefNum, &_interruptSocket, hidInterruptCallback, 0, btLibL2CapProtocol);
		BtLibSocketConnectInfoType connectInfo;
		connectInfo.remoteDeviceP = &gBdAddr;
		connectInfo.data.L2Cap.remotePsm = 0x13;
		connectInfo.data.L2Cap.minRemoteMtu = BT_L2CAP_MTU;
		connectInfo.data.L2Cap.localMtu = BT_L2CAP_MTU;
		error = BtLibSocketConnect(gBtLibRefNum, _interruptSocket, &connectInfo);
		if (error != btLibErrPending)
		{
			logPrintErr(error);
			return;
		}
	}
	else if (sEvent->event == btLibSocketEventSendComplete )
	{
		logPrintErr(sEvent->status);
		if (!bModeSet)
		{
			// Set Data report mode.
			UInt8 modeBuffer[] = {0x52, 0x12, 0x00, 0x31};
			Err error = BtLibSocketSend(gBtLibRefNum, _controlSocket, (UInt8*)&modeBuffer, sizeof(modeBuffer));
			logPrintErr(error);
			bModeSet = true;
		}
	}
}
//----------------------------------------------------------------

void hidInterruptCallback(BtLibSocketEventType *sEventP, UInt32 refCon)
{
	static UInt16 _buttonCache = 0x0000;
	static const WChar _keyMap[16] = {'2', '1', 'B', 'A',
	                                  '-', 0xFFFF, 0xFFFF,
	                                  'h', 'l', 'r',
	                                  'd', 'u', '+',
	                                  0xFFFF, 0xFFFF, 0xFFFF};

	if (sEventP->event == btLibSocketEventData)
	{
		UInt16 dataLength = sEventP->eventData.data.dataLen;
		UInt8* data = sEventP->eventData.data.data;

		// Parse mode 0x30 (core buttons)
		UInt16 buttonState, buttonStateChanges;

		logPrintf("%02x %02x %02x %02x", data[0], data[1], data[2], data[3]);

		if ((data[0] == 0xA1) && (data[1] == 0x30))
		{
			buttonState = (data[2]<<8) | data[3];
			buttonStateChanges = buttonState ^ _buttonCache;

			UInt16 bitMask = 1;
			UInt16 bitIndex;
			for (bitIndex=0 ; bitIndex<16 ; bitIndex++)
			{
				if (bitMask & buttonStateChanges)
				{
					if (bitMask & buttonState)
					{
						EventType downEvent = {keyDownEvent, false, 0, 0, 0, {0, 0, 0}};
						downEvent.data.keyDown.chr = _keyMap[bitIndex];
						//logPrintf("KeyDown: %C", _keyMap[bitIndex]);
						EvtAddEventToQueue(&downEvent);
					}
					else
					{
						EventType upEvent = {keyUpEvent, false, 0, 0, 0, {0, 0, 0}};
						upEvent.data.keyUp.chr = _keyMap[bitIndex];
						//logPrintf("KeyUp: %C", _keyMap[bitIndex]);
						EvtAddEventToQueue(&upEvent);
					}
				}
				bitMask <<= 1;
			}
		}

		// Save the previous button state.
		_buttonCache = buttonState;

		// Set LED pattern.
		static int led = 0x00;
		UInt8 ledBuffer[] = {0x52, 0x11, 0xa0};
		ledBuffer[2] = led;
		led += 0x10;
		BtLibSocketSend(gBtLibRefNum, _controlSocket, (UInt8*)&ledBuffer, sizeof(ledBuffer));
	}
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: btManagementCallback
//
// DESCRIPTION: Handle radio events.
//
// PARAMETERS:  See BtLib.h
//
// RETURNED: Nothing
//
//----------------------------------------------------------------
static void btManagementCallback(BtLibManagementEventType *mEventP, UInt32 refCon)
{
UNUSED(refCon);
Err error;

switch(mEventP->event)
    {
    case btLibManagementEventACLConnectOutbound:
        {
        logPrintf("MgmtEvent: ACLConnectOutbound");

        MemMove(&gBdAddr, &mEventP->eventData.bdAddr, sizeof(BtLibDeviceAddressType));
        BtLibAddrBtdToA(gBtLibRefNum, &mEventP->eventData.bdAddr, gBdAddrStr, kBdAddrStringSize);
        if(mEventP->status != btLibErrNoError)
            {
            logPrintf("Link to %s Failed", gBdAddrStr);
            return;
            }

        logPrintf("Link to %s Established", gBdAddrStr);
        ChangeLabel(MainAddrLabel, gBdAddrStr);
        
        // A host or device shall establish the control channel first, then the interrupt channel
        BtLibSocketCreate(gBtLibRefNum, &_controlSocket, hidControlCallback, 0, btLibL2CapProtocol);
        BtLibSocketConnectInfoType connectInfo;
        connectInfo.remoteDeviceP = &gBdAddr;
        connectInfo.data.L2Cap.remotePsm = 0x11;
        connectInfo.data.L2Cap.minRemoteMtu = BT_L2CAP_MTU;
        connectInfo.data.L2Cap.localMtu = BT_L2CAP_MTU;
        error = BtLibSocketConnect(gBtLibRefNum, _controlSocket, &connectInfo);
        if (error != btLibErrPending)
            {
            logPrintErr(error);
            break;
            }
        
        break;
        }

    case btLibManagementEventACLDisconnect:
        {
        logPrintf("MgmtEvent: ACLDisconnect");

        BtLibAddrBtdToA(gBtLibRefNum, &mEventP->eventData.bdAddr, gBdAddrStr, kBdAddrStringSize);
        logPrintf("Disconnected from %s, reason: 0x%hx", gBdAddrStr, mEventP->status);
        ChangeLabel(MainAddrLabel, "No Connection");
        MemSet(&gBdAddr, sizeof(gBdAddr), 0);
        break;
        }
    }
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: ChangeLabel
//
// DESCRIPTION: Utility function to change a label
// Note: The initial label (done in Constructor) MUST be
//  longer than any label you copy over it!!
//
// PARAMETERS:  label Id and new text
//
// RETURNED: Nothing
//
//----------------------------------------------------------------
static void ChangeLabel(UInt16 labelID, const Char *newLabel)
{
	FormPtr frm = FrmGetActiveForm();
	if (frm)
	{
		UInt16 labelIndex = FrmGetObjectIndex(frm, labelID);

		//Hide the label first to erase old text
		FrmHideObject(frm, labelIndex);
		//Copy in the new string
		FrmCopyLabel(frm, labelID, newLabel);
		//Now force the redraw
		FrmShowObject(frm, labelIndex);
	}
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: UpdateScrollbar
//
// DESCRIPTION: Update the scroll bar related to the log field.
//
// PARAMETERS:  None
//
// RETURNED: Nothing
//
//----------------------------------------------------------------
static void UpdateScrollbar()
{
	UInt16 currentPosition;
	UInt16 textHeigth;
	UInt16 fieldHeigth;
	UInt16   maxValue;
	FormPtr   frmP = FrmGetFormPtr(MainForm);
	if (frmP)
	{
		ScrollBarPtr scrollP = (ScrollBarPtr)FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainLogScrollBar));
		if (scrollP)
		{
			FieldPtr fldP = (FieldPtr)FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainLogField));

			FldGetScrollValues(fldP, &currentPosition, &textHeigth, &fieldHeigth);
			if (textHeigth > fieldHeigth)
				maxValue = textHeigth - fieldHeigth;
			else if (currentPosition)
				maxValue = currentPosition;
			else
				maxValue = 0;

			SclSetScrollBar(scrollP, (Int16)currentPosition, 0, (Int16)maxValue, (Int16)(fieldHeigth - 1));
		}
	}
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: ScrollLines
//
// DESCRIPTION: Scroll log field and update its scroll bar.
//
// PARAMETERS:  number of lines and direction
//  0 lines means a page scroll
//
// RETURNED: Nothing
//
//----------------------------------------------------------------
static void ScrollLines(UInt16 numLinesToScroll, WinDirectionType direction, Boolean update)
{
	FormPtr frmP = (FormPtr)FrmGetFormPtr(MainForm);
	FieldPtr fldP = (FieldPtr)FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainLogField));

	if (numLinesToScroll == 0)
		numLinesToScroll = FldGetVisibleLines(fldP) - 1;
	FldScrollField(fldP, numLinesToScroll, direction);
	if (update || (FldGetNumberOfBlankLines(fldP) && direction == winUp))
		UpdateScrollbar();
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: logPrintf
//
// DESCRIPTION: Add text to the log field, max 128 chars
//
// PARAMETERS:  same as StrPrintF
//
// RETURNED: nothing
//
//----------------------------------------------------------------
static void logPrintf(const Char *format, ...)
{
	static UInt16  gLenConv = 0;
	UInt16   lenMess;
	va_list   args;
	Char   mess[128];
	FormPtr frmP = FrmGetFormPtr(MainForm);
	if (!frmP)
		return;
	FieldPtr fldP = (FieldPtr)FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainLogField));
	if (!fldP)
		return;
	Char* txtP = FldGetTextPtr(fldP);

	// The first time, allocate the field with the maximum supported field size
	if (!txtP)
	{
		txtP = (Char*)MemPtrNew(kMaxSizeLog);
		gLenConv = FldGetVisibleLines(fldP);
		MemSet(txtP, gLenConv, '\n');
		txtP[gLenConv] = '\0';
	}

	// Format the message, if any, and append it to the field
	if (format)
	{
		// Format the message
		va_start(args, format);
		StrVPrintF(mess, format, args);
		va_end(args);

		// When field does not have enough room, keep only last half of the conversation
		// Move the last half in replacement of first half
		lenMess = (UInt16) StrLen(mess) + 1;
		if (gLenConv + lenMess >= kMaxSizeLog)
		{
			gLenConv >>= 1;
			MemMove(txtP, txtP + gLenConv + (gLenConv & 1), gLenConv );
		}

		// Append a carriage return and update length
		StrPrintF(txtP + gLenConv, "\n%s", mess);
		gLenConv += lenMess;
		FldSetTextPtr(fldP, txtP);
	}

	// If drawing is enabled (no menu or form is opened on top of the field)
	if (gCanDraw)
	{
		// Redraw field and scroll down to the bottom
		FldRecalculateField(fldP, false);
		FldScrollField(fldP, (UInt16) kMaxSizeLog, winDown);
		FldDrawField(fldP);
		UpdateScrollbar();
	}
}
//----------------------------------------------------------------

void logPrintErr(Err e)
{
	if (e == btLibErrNoError)
	{
		logPrintf("btLibErrNoError");
	}
else if (e == btLibErrError)
    {logPrintf("btLibErrError");}
else if (e == btLibErrNotOpen)
    {logPrintf("btLibErrNotOpen");}
else if (e == btLibErrBluetoothOff)
    {logPrintf("btLibErrBluetoothOff");}
else if (e == btLibErrNoPrefs)
    {logPrintf("btLibErrNoPrefs");}
else if (e == btLibErrAlreadyOpen)
    {logPrintf("btLibErrAlreadyOpen");}
else if (e == btLibErrOutOfMemory)
    {logPrintf("btLibErrOutOfMemory");}
else if (e == btLibErrFailed)
    {logPrintf("btLibErrFailed");}
else if (e == btLibErrInProgress)
    {logPrintf("btLibErrInProgress");}
else if (e == btLibErrParamError)
    {logPrintf("btLibErrParamError");}
else if (e == btLibErrTooMany)
    {logPrintf("btLibErrTooMany");}
else if (e == btLibErrPending)
    {logPrintf("btLibErrPending");}
else if (e == btLibErrNotInProgress)
    {logPrintf("btLibErrNotInProgress");}
else if (e == btLibErrRadioInitFailed)
    {logPrintf("btLibErrRadioInitFailed");}
else if (e == btLibErrRadioFatal)
    {logPrintf("btLibErrRadioFatal");}
else if (e == btLibErrRadioInitialized)
    {logPrintf("btLibErrRadioInitialized");}
else if (e == btLibErrRadioSleepWake)
    {logPrintf("btLibErrRadioSleepWake");}
else if (e == btLibErrNoConnection)
    {logPrintf("btLibErrNoConnection");}
else if (e == btLibErrAlreadyRegistered)
    {logPrintf("btLibErrAlreadyRegistered");}
else if (e == btLibErrNoAclLink)
    {logPrintf("btLibErrNoAclLink");}
else if (e == btLibErrSdpRemoteRecord)
    {logPrintf("btLibErrSdpRemoteRecord");}
else if (e == btLibErrSdpAdvertised)
    {logPrintf("btLibErrSdpAdvertised");}
else if (e == btLibErrSdpFormat)
    {logPrintf("btLibErrSdpFormat");}
else if (e == btLibErrSdpNotAdvertised)
    {logPrintf("btLibErrSdpNotAdvertised");}
else if (e == btLibErrSdpQueryVersion)
    {logPrintf("btLibErrSdpQueryVersion");}
else if (e == btLibErrSdpQueryHandle)
    {logPrintf("btLibErrSdpQueryHandle");}
else if (e == btLibErrSdpQuerySyntax)
    {logPrintf("btLibErrSdpQuerySyntax");}
else if (e == btLibErrSdpQueryPduSize)
    {logPrintf("btLibErrSdpQueryPduSize");}
else if (e == btLibErrSdpQueryContinuation)
    {logPrintf("btLibErrSdpQueryContinuation");}
else if (e == btLibErrSdpQueryResources)
    {logPrintf("btLibErrSdpQueryResources");}
else if (e == btLibErrSdpQueryDisconnect)
    {logPrintf("btLibErrSdpQueryDisconnect");}
else if (e == btLibErrSdpInvalidResponse)
    {logPrintf("btLibErrSdpInvalidResponse");}
else if (e == btLibErrSdpAttributeNotSet)
    {logPrintf("btLibErrSdpAttributeNotSet");}
else if (e == btLibErrSdpMapped)
    {logPrintf("btLibErrSdpMapped");}
else if (e == btLibErrSocket)
    {logPrintf("btLibErrSocket");}
else if (e == btLibErrSocketProtocol)
    {logPrintf("btLibErrSocketProtocol");}
else if (e == btLibErrSocketRole)
    {logPrintf("btLibErrSocketRole");}
else if (e == btLibErrSocketPsmUnavailable)
    {logPrintf("btLibErrSocketPsmUnavailable");}
else if (e == btLibErrSocketChannelUnavailable)
    {logPrintf("btLibErrSocketChannelUnavailable");}
else if (e == btLibErrSocketUserDisconnect)
    {logPrintf("btLibErrSocketUserDisconnect");}
else if (e == btLibErrCanceled)
    {logPrintf("btLibErrCanceled");}
else if (e == btLibErrBusy)
    {logPrintf("btLibErrBusy");}
else if (e == btLibErrInUseByService)
    {logPrintf("btLibErrInUseByService");}
else if (e == btLibErrNoPiconet)
    {logPrintf("btLibErrNoPiconet");}
else if (e == btLibErrRoleChange)
    {logPrintf("btLibErrRoleChange");}
else if (e == btLibErrSdpNotMapped)
    {logPrintf("btLibErrSdpNotMapped");}
else if (e == btLibErrAlreadyConnected)
    {logPrintf("btLibErrAlreadyConnected");}
else if (e == btLibErrStackNotOpen)
    {logPrintf("btLibErrStackNotOpen");}
else if (e == btLibErrBatteryTooLow)
    {logPrintf("btLibErrBatteryTooLow");}
else if (e == btLibErrNotFound)
    {logPrintf("btLibErrNotFound");}
else
    {logPrintf("Unknown error");}
}
//----------------------------------------------------------------

//----------------------------------------------------------------
//
// FUNCTION: PilotMain
//
// DESCRIPTION: This is the main entry point for the application.
//
// PARAMETERS:  cmd - word value specifying the launch code. 
//  cmdPB - pointer to a structure that is associated with the launch code. 
//  launchFlags -  word value providing extra information about the launch.
//
// RETURNED: Result of launch
//
//----------------------------------------------------------------
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	UInt32 btVersion;
	UNUSED(cmdPBP);

	if (FtrGet(btLibFeatureCreator, btLibFeatureVersion, &btVersion) != errNone)
	{
		// Alert the user if it's the active application
		if ((launchFlags & sysAppLaunchFlagNewGlobals) &&
			(launchFlags & sysAppLaunchFlagUIApp))
		{
			FrmAlert(MissingBtComponentsAlert);
		}

		return (sysErrRomIncompatible);
	}

	if (cmd == sysAppLaunchCmdNormalLaunch)
	{
		AppStart();
		AppEventLoop();
		AppStop();
	}

	return (errNone);
}
//----------------------------------------------------------------
